@@ -1,32 +1,31 @@
 /*
 
-  sftp_client.c 
+  sftp_client.c
 
   Author: Pekka Riikonen <priikone@silcnet.org>
 
-  Copyright (C) 2001 Pekka Riikonen
+  Copyright (C) 2001 - 2007 Pekka Riikonen
 
-  This program is free software; you can redistribute it and/or modify
-  it under the terms of the GNU General Public License as published by
-  the Free Software Foundation; version 2 of the License.
+  The contents of this file are subject to one of the Licenses specified 
+  in the COPYING file;  You may not use this file except in compliance 
+  with the License.
 
-  This program is distributed in the hope that it will be useful,
-  but WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-  GNU General Public License for more details.
+  The software distributed under the License is distributed on an "AS IS"
+  basis, in the hope that it will be useful, but WITHOUT WARRANTY OF ANY
+  KIND, either expressed or implied.  See the COPYING file for more
+  information.
 
 */
-/* $Id: sftp_client.c,v 1.8 2002/06/14 20:33:49 priikone Exp $ */
+/* $Id: sftp_client.c,v 1.17.2.1 2007/06/06 14:31:13 priikone Exp $ */
 
-#include "silcincludes.h"
+#include "silc.h"
 #include "silcsftp.h"
 #include "sftp_util.h"
 
 /* Request context. Every request will allocate this context and set
    the correct callback function according the `type' field. */
 typedef struct SilcSFTPRequestStruct {
-  SilcUInt32 id;
-  SilcSFTPPacket type;
+  struct SilcSFTPRequestStruct *next;
   SilcSFTPStatusCallback status;
   SilcSFTPHandleCallback handle;
   SilcSFTPDataCallback data;
@@ -34,18 +33,20 @@ typedef struct SilcSFTPRequestStruct {
   SilcSFTPAttrCallback attr;
   SilcSFTPExtendedCallback extended;
   void *context;
-  struct SilcSFTPRequestStruct *next;
+  SilcUInt32 id;
+  SilcSFTPPacket type;
 } *SilcSFTPRequest;
 
 /* SFTP client context */
 typedef struct {
-  SilcSFTPSendPacketCallback send_packet;
-  void *send_context;
+  SilcStream stream;
+  SilcSchedule schedule;
   SilcSFTPVersionCallback version;
-  void *version_context;
-  SilcUInt32 id;
+  SilcSFTPErrorCallback error;
+  void *context;
   SilcList requests;
   SilcBuffer packet;
+  SilcUInt32 id;
 } *SilcSFTPClient;
 
 /* File handle */
@@ -54,6 +55,9 @@ struct SilcSFTPHandleStruct {
   SilcUInt32 data_len;
 };
 
+static void silc_sftp_client_receive_process(SilcSFTP context,
+					     SilcBuffer buffer);
+
 /* Creates SilcSFTPHandle and returns pointer to it. The caller must free
    the context. */
 
@@ -84,7 +88,7 @@ static void silc_sftp_handle_delete(SilcSFTPHandle handle)
 
 /* Returns the handle data of the `handle' to the `data' pointer. */
 
-static void silc_sftp_handle_get(SilcSFTPHandle handle, 
+static void silc_sftp_handle_get(SilcSFTPHandle handle,
 				 const unsigned char **data,
 				 SilcUInt32 *data_len)
 {
@@ -92,14 +96,15 @@ static void silc_sftp_handle_get(SilcSFTPHandle handle,
   *data_len = handle->data_len;
 }
 
-/* General routine to send SFTP packet to the SFTP server. */
+/* Generic routine to send SFTP packet to the SFTP server. */
 
 static void silc_sftp_send_packet(SilcSFTPClient sftp,
-				  SilcSFTPPacket type, 
+				  SilcSFTPPacket type,
 				  SilcUInt32 len, ...)
 {
   SilcBuffer tmp;
   va_list vp;
+  int ret;
 
   va_start(vp, len);
   tmp = silc_sftp_packet_encode_vp(type, sftp->packet, len, vp);
@@ -108,20 +113,39 @@ static void silc_sftp_send_packet(SilcSFTPClient sftp,
     return;
   sftp->packet = tmp;
 
-  SILC_LOG_HEXDUMP(("SFTP packet to server"), sftp->packet->data, 
-		   sftp->packet->len);
+  SILC_LOG_HEXDUMP(("SFTP packet to server"), sftp->packet->data,
+		   silc_buffer_len(sftp->packet));
 
   /* Send the packet */
-  (*sftp->send_packet)(sftp->packet, sftp->send_context);
+  while (silc_buffer_len(sftp->packet) > 0) {
+    ret = silc_stream_write(sftp->stream, silc_buffer_data(sftp->packet),
+			    silc_buffer_len(sftp->packet));
+    if (ret == -2) {
+      SILC_LOG_ERROR(("Error sending SFTP packet type %d", type));
+      sftp->error((SilcSFTP)sftp, SILC_SFTP_STATUS_NO_CONNECTION,
+		  sftp->context);
+      silc_buffer_reset(sftp->packet);
+      return;
+    }
+    if (ret == 0) {
+      sftp->error((SilcSFTP)sftp, SILC_SFTP_STATUS_EOF, sftp->context);
+      silc_buffer_reset(sftp->packet);
+      return;
+    }
+    if (ret == -1)
+      return;
+
+    silc_buffer_pull(sftp->packet, ret);
+  }
 
   /* Clear packet */
-  sftp->packet->data = sftp->packet->tail = sftp->packet->head;
-  sftp->packet->len = 0;
+  silc_buffer_reset(sftp->packet);
 }
 
 /* Finds request by request ID. */
 
-static SilcSFTPRequest silc_sftp_find_request(SilcSFTPClient sftp, SilcUInt32 id)
+static SilcSFTPRequest silc_sftp_find_request(SilcSFTPClient sftp,
+					      SilcUInt32 id)
 {
   SilcSFTPRequest req;
 
@@ -133,7 +157,7 @@ static SilcSFTPRequest silc_sftp_find_request(SilcSFTPClient sftp, SilcUInt32 id
       return req;
   }
 
-  SILC_LOG_DEBUG(("Unknown request ID"));
+  SILC_LOG_DEBUG(("Unknown request ID %d", id));
 
   return NULL;
 }
@@ -143,8 +167,8 @@ static SilcSFTPRequest silc_sftp_find_request(SilcSFTPClient sftp, SilcUInt32 id
    operation. The variable argument list includes the status and req->type
    specific data. */
 
-static void silc_sftp_call_request(SilcSFTPClient sftp, 
-				   SilcSFTPRequest req, 
+static void silc_sftp_call_request(SilcSFTPClient sftp,
+				   SilcSFTPRequest req,
 				   SilcSFTPPacket type,
 				   SilcSFTPStatus status, ...)
 {
@@ -171,9 +195,9 @@ static void silc_sftp_call_request(SilcSFTPClient sftp,
       data_len = (SilcUInt32)va_arg(vp, SilcUInt32);
 
       if (req->data)
-	(*req->data)((SilcSFTP)sftp, status, (const unsigned char *)data, 
+	(*req->data)((SilcSFTP)sftp, status, (const unsigned char *)data,
 		     data_len, req->context);
-    }    
+    }
     break;
 
   case SILC_SFTP_OPEN:
@@ -221,7 +245,7 @@ static void silc_sftp_call_request(SilcSFTPClient sftp,
       language_tag = (char *)va_arg(vp, char *);
 
       if (req->status)
-	(*req->status)((SilcSFTP)sftp, status, (const char *)message, 
+	(*req->status)((SilcSFTP)sftp, status, (const char *)message,
 		       (const char *)language_tag, req->context);
     }
     break;
@@ -242,7 +266,7 @@ static void silc_sftp_call_request(SilcSFTPClient sftp,
       attr = (SilcSFTPAttributes)va_arg(vp, SilcSFTPAttributes);
 
       if (req->attr)
-	(*req->attr)((SilcSFTP)sftp, status, (const SilcSFTPAttributes)attr, 
+	(*req->attr)((SilcSFTP)sftp, status, (const SilcSFTPAttributes)attr,
 		     req->context);
     }
     break;
@@ -283,12 +307,13 @@ static void silc_sftp_call_request(SilcSFTPClient sftp,
       data_len = (SilcUInt32)va_arg(vp, SilcUInt32);
 
       if (req->extended)
-	(*req->extended)((SilcSFTP)sftp, status, (const unsigned char *)data, 
+	(*req->extended)((SilcSFTP)sftp, status, (const unsigned char *)data,
 			 data_len, req->context);
     }
     break;
 
   default:
+    SILC_LOG_DEBUG(("Unknown request type %d", req->type));
     break;
   }
 
@@ -299,33 +324,103 @@ static void silc_sftp_call_request(SilcSFTPClient sftp,
   va_end(vp);
 }
 
-/* Starts SFTP client and returns context for it.  The version callback
-   indicated by the `callback' will be called after the SFTP session has
-   been started and server has returned the version of the protocol.  The
-   SFTP client context is returned in the callback too.  This returns 
-   allocated SFTP client context or NULL on error. */
+/* Handles stream I/O */
 
-SilcSFTP silc_sftp_client_start(SilcSFTPSendPacketCallback send_packet,
-				void *send_context,
-				SilcSFTPVersionCallback callback,
+static void silc_sftp_client_io(SilcStream stream, SilcStreamStatus status,
+				void *context)
+{
+  SilcSFTPClient sftp = context;
+  unsigned char inbuf[65536];
+  SilcBufferStruct packet;
+  int ret;
+
+  switch (status) {
+  case SILC_STREAM_CAN_READ:
+    SILC_LOG_DEBUG(("Reading data from stream"));
+
+    /* Read data from stream */
+    ret = silc_stream_read(stream, inbuf, sizeof(inbuf));
+    if (ret <= 0) {
+      if (ret == 0)
+	sftp->error(context, SILC_SFTP_STATUS_EOF, sftp->context);
+      if (ret == -2)
+	sftp->error(context, SILC_SFTP_STATUS_NO_CONNECTION, sftp->context);
+      return;
+    }
+
+    SILC_LOG_DEBUG(("Read %d bytes", ret));
+
+    /* Now process the SFTP packet */
+    silc_buffer_set(&packet, inbuf, ret);
+    silc_sftp_client_receive_process(context, &packet);
+    break;
+
+  case SILC_STREAM_CAN_WRITE:
+    if (!silc_buffer_headlen(sftp->packet))
+      return;
+
+    SILC_LOG_DEBUG(("Writing pending data to stream"));
+
+    /* Write pending data to stream */
+    silc_buffer_push(sftp->packet, silc_buffer_headlen(sftp->packet));
+    while (silc_buffer_len(sftp->packet) > 0) {
+      ret = silc_stream_write(stream, sftp->packet->data,
+			      silc_buffer_len(sftp->packet));
+      if (ret == 0) {
+	sftp->error(context, SILC_SFTP_STATUS_EOF, sftp->context);
+	silc_buffer_reset(sftp->packet);
+	return;
+      }
+
+      if (ret == -2) {
+	sftp->error(context, SILC_SFTP_STATUS_NO_CONNECTION, sftp->context);
+	silc_buffer_reset(sftp->packet);
+	return;
+      }
+
+      if (ret == -1)
+	return;
+
+      /* Wrote data */
+      silc_buffer_pull(sftp->packet, ret);
+    }
+    break;
+
+  default:
+    break;
+  }
+}
+
+/* Starts SFTP client and returns context for it. */
+
+SilcSFTP silc_sftp_client_start(SilcStream stream,
+				SilcSchedule schedule,
+				SilcSFTPVersionCallback version_cb,
+				SilcSFTPErrorCallback error_cb,
 				void *context)
 {
   SilcSFTPClient sftp;
 
-  if (!send_packet)
+  SILC_LOG_DEBUG(("Starting SFTP client"));
+
+  if (!stream)
     return NULL;
 
   sftp = silc_calloc(1, sizeof(*sftp));
   if (!sftp)
     return NULL;
-  sftp->send_packet = send_packet;
-  sftp->send_context = send_context;
-  sftp->version = callback;
-  sftp->version_context = context;
+  sftp->stream = stream;
+  sftp->version = version_cb;
+  sftp->error = error_cb;
+  sftp->context = context;
+  sftp->schedule = schedule;
   silc_list_init(sftp->requests, struct SilcSFTPRequestStruct, next);
 
+  /* We handle the stream now */
+  silc_stream_set_notifier(stream, schedule, silc_sftp_client_io, sftp);
+
   /* Send the SFTP session initialization to the server */
-  silc_sftp_send_packet(sftp, SILC_SFTP_INIT, 4, 
+  silc_sftp_send_packet(sftp, SILC_SFTP_INIT, 4,
 			SILC_STR_UI_INT(SILC_SFTP_PROTOCOL_VERSION),
 			SILC_STR_END);
 
@@ -340,39 +435,60 @@ void silc_sftp_client_shutdown(SilcSFTP context)
 {
   SilcSFTPClient sftp = (SilcSFTPClient)context;
 
+  silc_stream_set_notifier(sftp->stream, sftp->schedule, NULL, NULL);
   if (sftp->packet)
     silc_buffer_free(sftp->packet);
   silc_free(sftp);
 }
 
 /* Function that is called to process the incmoing SFTP packet. */
-/* XXX Some day this will go away and we have automatic receive callbacks
-   for SilcSocketConnection API or SilcPacketContext API. */
 
-void silc_sftp_client_receive_process(SilcSFTP context,
-				      SilcSocketConnection sock,
-				      SilcPacketContext *packet)
+void silc_sftp_client_receive_process(SilcSFTP context, SilcBuffer buffer)
 {
   SilcSFTPClient sftp = (SilcSFTPClient)context;
   SilcSFTPRequest req;
   SilcSFTPPacket type;
-  const unsigned char *payload = NULL;
+  unsigned char *payload = NULL;
   SilcUInt32 payload_len;
   int ret;
   SilcBufferStruct buf;
   SilcUInt32 id;
 
-  SILC_LOG_DEBUG(("Start"));
+  SILC_LOG_DEBUG(("Process SFTP packet"));
 
   /* Parse the packet */
-  type = silc_sftp_packet_decode(packet->buffer, (unsigned char **)&payload, 
-				 &payload_len);
-  if (!type)
+  type = silc_sftp_packet_decode(buffer, &payload, &payload_len);
+  if (type <= 0)
     return;
 
-  silc_buffer_set(&buf, (unsigned char *)payload, payload_len);
+  silc_buffer_set(&buf, payload, payload_len);
 
   switch (type) {
+  case SILC_SFTP_DATA:
+    {
+      unsigned char *data = NULL;
+      SilcUInt32 data_len = 0;
+
+      SILC_LOG_DEBUG(("Data packet"));
+
+      ret = silc_buffer_unformat(&buf,
+				 SILC_STR_UI_INT(&id),
+				 SILC_STR_UI32_NSTRING(&data, &data_len),
+				 SILC_STR_END);
+      if (ret < 0)
+	break;
+
+      /* Get request */
+      req = silc_sftp_find_request(sftp, id);
+      if (!req)
+	break;
+
+      /* Call the callback */
+      silc_sftp_call_request(sftp, req, type, SILC_SFTP_STATUS_OK,
+			     data, data_len);
+    }
+    break;
+
   case SILC_SFTP_VERSION:
     {
       SilcSFTPVersion version;
@@ -383,14 +499,14 @@ void silc_sftp_client_receive_process(SilcSFTP context,
 				 SILC_STR_UI_INT(&version),
 				 SILC_STR_END);
       if (ret < 0) {
-	(*sftp->version)((SilcSFTP)sftp, SILC_SFTP_STATUS_FAILURE, 0, 
-			 sftp->version_context);
+	(*sftp->version)((SilcSFTP)sftp, SILC_SFTP_STATUS_FAILURE, 0,
+			 sftp->context);
 	break;
       }
 
       /* Call the callback */
-      (*sftp->version)((SilcSFTP)sftp, SILC_SFTP_STATUS_OK, version, 
-		       sftp->version_context);
+      (*sftp->version)((SilcSFTP)sftp, SILC_SFTP_STATUS_OK, version,
+		       sftp->context);
     }
     break;
 
@@ -401,7 +517,7 @@ void silc_sftp_client_receive_process(SilcSFTP context,
 
       SILC_LOG_DEBUG(("Status packet"));
 
-      ret = silc_buffer_unformat(&buf, 
+      ret = silc_buffer_unformat(&buf,
 				 SILC_STR_UI_INT(&id),
 				 SILC_STR_UI_INT(&status),
 				 SILC_STR_END);
@@ -443,9 +559,9 @@ void silc_sftp_client_receive_process(SilcSFTP context,
 
       SILC_LOG_DEBUG(("Handle packet"));
 
-      ret = silc_buffer_unformat(&buf, 
+      ret = silc_buffer_unformat(&buf,
 				 SILC_STR_UI_INT(&id),
-				 SILC_STR_UI32_NSTRING(&handle, 
+				 SILC_STR_UI32_NSTRING(&handle,
 						       &handle_len),
 				 SILC_STR_END);
       if (ret < 0)
@@ -457,36 +573,11 @@ void silc_sftp_client_receive_process(SilcSFTP context,
 	break;
 
       /* Call the callback */
-      silc_sftp_call_request(sftp, req, type, SILC_SFTP_STATUS_OK, 
+      silc_sftp_call_request(sftp, req, type, SILC_SFTP_STATUS_OK,
 			     handle, handle_len);
     }
     break;
 
-  case SILC_SFTP_DATA:
-    {
-      unsigned char *data = NULL;
-      SilcUInt32 data_len = 0;
-
-      SILC_LOG_DEBUG(("Data packet"));
-
-      ret = silc_buffer_unformat(&buf, 
-				 SILC_STR_UI_INT(&id),
-				 SILC_STR_UI32_NSTRING(&data, &data_len),
-				 SILC_STR_END);
-      if (ret < 0)
-	break;
-
-      /* Get request */
-      req = silc_sftp_find_request(sftp, id);
-      if (!req)
-	break;
-
-      /* Call the callback */
-      silc_sftp_call_request(sftp, req, type, SILC_SFTP_STATUS_OK, 
-			     data, data_len);
-    }
-    break;
-
   case SILC_SFTP_NAME:
     {
       SilcUInt32 count;
@@ -494,7 +585,7 @@ void silc_sftp_client_receive_process(SilcSFTP context,
 
       SILC_LOG_DEBUG(("Name packet"));
 
-      ret = silc_buffer_unformat(&buf, 
+      ret = silc_buffer_unformat(&buf,
 				 SILC_STR_UI_INT(&id),
 				 SILC_STR_UI_INT(&count),
 				 SILC_STR_END);
@@ -526,10 +617,11 @@ void silc_sftp_client_receive_process(SilcSFTP context,
 
       SILC_LOG_DEBUG(("Attributes packet"));
 
-      ret = silc_buffer_unformat(&buf, 
-				 SILC_STR_UI_INT(&id),
-				 SILC_STR_UI_XNSTRING(&data, buf.len - 4),
-				 SILC_STR_END);
+      ret =
+	silc_buffer_unformat(&buf,
+			     SILC_STR_UI_INT(&id),
+			     SILC_STR_DATA(&data, silc_buffer_len(&buf) - 4),
+			     SILC_STR_END);
       if (ret < 0)
 	break;
 
@@ -538,7 +630,7 @@ void silc_sftp_client_receive_process(SilcSFTP context,
       if (!req)
 	break;
 
-      silc_buffer_set(&tmpbuf, data, buf.len - 4);
+      silc_buffer_set(&tmpbuf, data, silc_buffer_len(&buf) - 4);
       attr = silc_sftp_attr_decode(&tmpbuf);
       if (!attr)
 	break;
@@ -554,9 +646,10 @@ void silc_sftp_client_receive_process(SilcSFTP context,
 
       SILC_LOG_DEBUG(("Extended reply packet"));
 
-      ret = silc_buffer_unformat(&buf, 
+      ret = silc_buffer_unformat(&buf,
 				 SILC_STR_UI_INT(&id),
-				 SILC_STR_UI_XNSTRING(&data, buf.len - 4),
+				 SILC_STR_DATA(&data,
+					       silc_buffer_len(&buf) - 4),
 				 SILC_STR_END);
       if (ret < 0)
 	break;
@@ -567,8 +660,8 @@ void silc_sftp_client_receive_process(SilcSFTP context,
 	break;
 
       /* Call the callback */
-      silc_sftp_call_request(sftp, req, type, SILC_SFTP_STATUS_OK, 
-			     data, buf.len - 4);
+      silc_sftp_call_request(sftp, req, type, SILC_SFTP_STATUS_OK,
+			     data, silc_buffer_len(&buf) - 4);
     }
     break;
 
@@ -577,7 +670,7 @@ void silc_sftp_client_receive_process(SilcSFTP context,
   }
 }
 
-void silc_sftp_open(SilcSFTP sftp, 
+void silc_sftp_open(SilcSFTP sftp,
 		    const char *filename,
 		    SilcSFTPFileOperation pflags,
 		    SilcSFTPAttributes attrs,
@@ -603,15 +696,15 @@ void silc_sftp_open(SilcSFTP sftp,
   attrs_buf = silc_sftp_attr_encode(attrs);
   if (!attrs_buf)
     return;
-  len = 4 + 4 + strlen(filename) + 4 + attrs_buf->len;
+  len = 4 + 4 + strlen(filename) + 4 + silc_buffer_len(attrs_buf);
 
-  silc_sftp_send_packet(client, req->type, len, 
+  silc_sftp_send_packet(client, req->type, len,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(strlen(filename)),
 			SILC_STR_UI32_STRING(filename),
 			SILC_STR_UI_INT(pflags),
-			SILC_STR_UI_XNSTRING(attrs_buf->data, 
-					     attrs_buf->len),
+			SILC_STR_UI_XNSTRING(attrs_buf->data,
+					     silc_buffer_len(attrs_buf)),
 			SILC_STR_END);
 
   silc_buffer_free(attrs_buf);
@@ -642,16 +735,17 @@ void silc_sftp_close(SilcSFTP sftp,
   silc_sftp_handle_get(handle, &hdata, &hdata_len);
   len = 4 + 4 + hdata_len;
 
-  silc_sftp_send_packet(client, req->type, len, 
+  silc_sftp_send_packet(client, req->type, len,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(hdata_len),
 			SILC_STR_UI_XNSTRING(hdata, hdata_len),
 			SILC_STR_END);
+  silc_sftp_handle_delete(handle);
 }
 
 void silc_sftp_read(SilcSFTP sftp,
 		    SilcSFTPHandle handle,
-		    SilcUInt64 offset, 
+		    SilcUInt64 offset,
 		    SilcUInt32 len,
 		    SilcSFTPDataCallback callback,
 		    void *context)
@@ -676,7 +770,7 @@ void silc_sftp_read(SilcSFTP sftp,
   silc_sftp_handle_get(handle, &hdata, &hdata_len);
   len2 = 4 + 4 + hdata_len + 8 + 4;
 
-  silc_sftp_send_packet(client, req->type, len2, 
+  silc_sftp_send_packet(client, req->type, len2,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(hdata_len),
 			SILC_STR_UI_XNSTRING(hdata, hdata_len),
@@ -713,7 +807,7 @@ void silc_sftp_write(SilcSFTP sftp,
   silc_sftp_handle_get(handle, &hdata, &hdata_len);
   len = 4 + 4 + hdata_len + 8 + 4 + data_len;
 
-  silc_sftp_send_packet(client, req->type, len, 
+  silc_sftp_send_packet(client, req->type, len,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(hdata_len),
 			SILC_STR_UI_XNSTRING(hdata, hdata_len),
@@ -745,7 +839,7 @@ void silc_sftp_remove(SilcSFTP sftp,
 
   len = 4 + 4 + strlen(filename);
 
-  silc_sftp_send_packet(client, req->type, len, 
+  silc_sftp_send_packet(client, req->type, len,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(strlen(filename)),
 			SILC_STR_UI32_STRING(filename),
@@ -775,7 +869,7 @@ void silc_sftp_rename(SilcSFTP sftp,
 
   len = 4 + 4 + strlen(oldname) + 4 + strlen(newname);
 
-  silc_sftp_send_packet(client, req->type, len, 
+  silc_sftp_send_packet(client, req->type, len,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(strlen(oldname)),
 			SILC_STR_UI32_STRING(oldname),
@@ -809,14 +903,14 @@ void silc_sftp_mkdir(SilcSFTP sftp,
   attrs_buf = silc_sftp_attr_encode(attrs);
   if (!attrs_buf)
     return;
-  len = 4 + 4 + strlen(path) + attrs_buf->len;
+  len = 4 + 4 + strlen(path) + silc_buffer_len(attrs_buf);
 
-  silc_sftp_send_packet(client, req->type, len, 
+  silc_sftp_send_packet(client, req->type, len,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(strlen(path)),
 			SILC_STR_UI32_STRING(path),
 			SILC_STR_UI_XNSTRING(attrs_buf->data,
-					     attrs_buf->len),
+					     silc_buffer_len(attrs_buf)),
 			SILC_STR_END);
 
   silc_buffer_free(attrs_buf);
@@ -844,7 +938,7 @@ void silc_sftp_rmdir(SilcSFTP sftp,
 
   len = 4 + 4 + strlen(path);
 
-  silc_sftp_send_packet(client, req->type, len, 
+  silc_sftp_send_packet(client, req->type, len,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(strlen(path)),
 			SILC_STR_UI32_STRING(path),
@@ -873,7 +967,7 @@ void silc_sftp_opendir(SilcSFTP sftp,
 
   len = 4 + 4 + strlen(path);
 
-  silc_sftp_send_packet(client, req->type, len, 
+  silc_sftp_send_packet(client, req->type, len,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(strlen(path)),
 			SILC_STR_UI32_STRING(path),
@@ -905,7 +999,7 @@ void silc_sftp_readdir(SilcSFTP sftp,
   silc_sftp_handle_get(handle, &hdata, &hdata_len);
   len = 4 + 4 + hdata_len;
 
-  silc_sftp_send_packet(client, req->type, len, 
+  silc_sftp_send_packet(client, req->type, len,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(hdata_len),
 			SILC_STR_UI_XNSTRING(hdata, hdata_len),
@@ -934,7 +1028,7 @@ void silc_sftp_stat(SilcSFTP sftp,
 
   len = 4 + 4 + strlen(path);
 
-  silc_sftp_send_packet(client, req->type, len, 
+  silc_sftp_send_packet(client, req->type, len,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(strlen(path)),
 			SILC_STR_UI32_STRING(path),
@@ -963,7 +1057,7 @@ void silc_sftp_lstat(SilcSFTP sftp,
 
   len = 4 + 4 + strlen(path);
 
-  silc_sftp_send_packet(client, req->type, len, 
+  silc_sftp_send_packet(client, req->type, len,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(strlen(path)),
 			SILC_STR_UI32_STRING(path),
@@ -995,7 +1089,7 @@ void silc_sftp_fstat(SilcSFTP sftp,
   silc_sftp_handle_get(handle, &hdata, &hdata_len);
   len = 4 + 4 + hdata_len;
 
-  silc_sftp_send_packet(client, req->type, len, 
+  silc_sftp_send_packet(client, req->type, len,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(hdata_len),
 			SILC_STR_UI_XNSTRING(hdata, hdata_len),
@@ -1027,14 +1121,14 @@ void silc_sftp_setstat(SilcSFTP sftp,
   attrs_buf = silc_sftp_attr_encode(attrs);
   if (!attrs_buf)
     return;
-  len = 4 + 4 + strlen(path) + attrs_buf->len;
+  len = 4 + 4 + strlen(path) + silc_buffer_len(attrs_buf);
 
-  silc_sftp_send_packet(client, req->type, len, 
+  silc_sftp_send_packet(client, req->type, len,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(strlen(path)),
 			SILC_STR_UI32_STRING(path),
 			SILC_STR_UI_XNSTRING(attrs_buf->data,
-					     attrs_buf->len),
+					     silc_buffer_len(attrs_buf)),
 			SILC_STR_END);
 
   silc_buffer_free(attrs_buf);
@@ -1068,14 +1162,14 @@ void silc_sftp_fsetstat(SilcSFTP sftp,
   attrs_buf = silc_sftp_attr_encode(attrs);
   if (!attrs_buf)
     return;
-  len = 4 + 4 + hdata_len + attrs_buf->len;
+  len = 4 + 4 + hdata_len + silc_buffer_len(attrs_buf);
 
-  silc_sftp_send_packet(client, req->type, len, 
+  silc_sftp_send_packet(client, req->type, len,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(hdata_len),
 			SILC_STR_UI_XNSTRING(hdata, hdata_len),
 			SILC_STR_UI_XNSTRING(attrs_buf->data,
-					     attrs_buf->len),
+					     silc_buffer_len(attrs_buf)),
 			SILC_STR_END);
 
   silc_buffer_free(attrs_buf);
@@ -1103,7 +1197,7 @@ void silc_sftp_readlink(SilcSFTP sftp,
 
   len = 4 + 4 + strlen(path);
 
-  silc_sftp_send_packet(client, req->type, len, 
+  silc_sftp_send_packet(client, req->type, len,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(strlen(path)),
 			SILC_STR_UI32_STRING(path),
@@ -1133,7 +1227,7 @@ void silc_sftp_symlink(SilcSFTP sftp,
 
   len = 4 + 4 + strlen(linkpath) + 4 + strlen(targetpath);
 
-  silc_sftp_send_packet(client, req->type, len, 
+  silc_sftp_send_packet(client, req->type, len,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(strlen(linkpath)),
 			SILC_STR_UI32_STRING(linkpath),
@@ -1164,7 +1258,7 @@ void silc_sftp_realpath(SilcSFTP sftp,
 
   len = 4 + 4 + strlen(path);
 
-  silc_sftp_send_packet(client, req->type, len, 
+  silc_sftp_send_packet(client, req->type, len,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(strlen(path)),
 			SILC_STR_UI32_STRING(path),
@@ -1195,7 +1289,7 @@ void silc_sftp_extended(SilcSFTP sftp,
 
   len = 4 + 4 + strlen(request) + data_len;
 
-  silc_sftp_send_packet(client, req->type, len, 
+  silc_sftp_send_packet(client, req->type, len,
 			SILC_STR_UI_INT(req->id),
 			SILC_STR_UI_INT(strlen(request)),
 			SILC_STR_UI32_STRING(request),
