@@ -16,7 +16,7 @@
   information.
 
 */
-/* $Id: silcunixschedule.c,v 1.28.2.5 2007/08/07 18:15:19 priikone Exp $ */
+/* $Id: silcunixschedule.c,v 1.28.2.6 2007/12/30 12:04:28 priikone Exp $ */
 
 #include "silc.h"
 
@@ -498,6 +498,7 @@ void silc_schedule_internal_signal_register(SilcSchedule schedule,
       signal_call[i].sig = sig;
       signal_call[i].callback = callback;
       signal_call[i].context = callback_context;
+      signal_call[i].schedule = schedule;
       signal_call[i].call = FALSE;
       signal(sig, silc_schedule_internal_sighandler);
       break;
@@ -527,6 +528,7 @@ void silc_schedule_internal_signal_unregister(SilcSchedule schedule,
       signal_call[i].sig = 0;
       signal_call[i].callback = NULL;
       signal_call[i].context = NULL;
+      signal_call[i].schedule = NULL;
       signal_call[i].call = FALSE;
       signal(sig, SIG_DFL);
     }
